@@ -113,7 +113,7 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 		        USER_OBJECT_ fullNamespaceInfo, USER_OBJECT_ r_encoding,
 		        USER_OBJECT_ useDotNames,
       		         USER_OBJECT_ xinclude, USER_OBJECT_ errorFun,
-                         USER_OBJECT_ manageMemory)
+		           USER_OBJECT_ manageMemory, USER_OBJECT_ r_parserOptions)
 {
 
   const char *name;
@@ -129,9 +129,16 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 
   const char *encoding = NULL;
   int freeName = 0;
+  int parserOptions = 0;
 
-  if(GET_LENGTH(r_encoding))
+  if(GET_LENGTH(r_encoding)) {
       encoding = CHAR(STRING_ELT(r_encoding, 0));
+      if(!encoding[0])
+	  encoding = NULL;
+  }
+
+  if(Rf_length(r_parserOptions))
+     parserOptions = INTEGER(r_parserOptions)[0];
 
   parserSettings.skipBlankLines = LOGICAL_DATA(skipBlankLines)[0];
   parserSettings.converters = converterFunctions;
@@ -189,13 +196,16 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 #endif
 
   if(asTextBuffer) {
-      doc = useHTML ? htmlParseDoc(CHAR_TO_XMLCHAR(name), encoding) : xmlParseMemory(name, strlen(name)); 
+      doc = useHTML ? htmlParseDoc(CHAR_TO_XMLCHAR(name), encoding) : 
+	              xmlReadMemory(name, strlen(name), NULL, encoding, parserOptions) ;
+                	  /* xmlParseMemory(name, strlen(name)) */ 
 
       if(doc != NULL) 
          doc->name = (char *) xmlStrdup(CHAR_TO_XMLCHAR("<buffer>"));
 
   } else {
-      doc = useHTML ? htmlParseFile(XMLCHAR_TO_CHAR(name), encoding) : xmlParseFile(name);
+      doc = useHTML ? htmlParseFile(XMLCHAR_TO_CHAR(name), encoding) : 
+	              xmlReadFile(name, encoding, parserOptions) /* xmlParseFile(name) */ ;
   }
 
 #ifdef RS_XML_SET_STRUCTURED_ERROR 
@@ -1252,11 +1262,13 @@ makeSchemaReference(xmlSchemaPtr schema)
 }
 
 
+#ifndef NO_XML_MEMORY_SHOW_ROUTINE
 void
 RS_XML_MemoryShow()
 {
     xmlMemDisplay(stderr);
 }
+#endif
 
 USER_OBJECT_
 RS_XML_setDocumentName(USER_OBJECT_ sdoc, USER_OBJECT_ sname)
@@ -1489,6 +1501,24 @@ R_getDocEncoding(SEXP r_doc)
 }
 
 
+int 
+getTextElementLineNumber(xmlNodePtr node)
+{
+    int val = -1;
+
+    if(node->parent)
+	val = node->parent->line;
+
+    xmlNodePtr prev = node->prev;
+    while(prev) {
+	if(prev->line > 0) {
+	    val = prev->line;
+	    break;
+	}
+	prev = prev->prev;
+    }
+    return(val);
+}
 
 SEXP
 R_getLineNumber(SEXP r_node)
@@ -1499,7 +1529,9 @@ R_getLineNumber(SEXP r_node)
 	return(NEW_INTEGER(0));
     }
 
-    return(ScalarInteger(node->line));
+//    XML_GET_LINE(node)
+    return(ScalarInteger(node->line == 0 ? 
+			 getTextElementLineNumber(node) : node->line));
 }
 
 
